from typing import NamedTuple

from tensorflow.keras import Model
from tensorflow.keras.layers import Layer
from tensorflow.keras.layers import Concatenate

class Concatenation(NamedTuple):
    layer_one: str
    layer_two: str
    axis: int
    name: str

# SequentialWithConcatenations([
#   Input((32, 32, 1)),
#   Conv2D(..., name="layer1"),
#   Conv2D(..., name="layer2"),
#   Conv2D(..., name="layer3"),
#   Concatenation("layer2", "layer3", 3, "layer4")
#   Conv2D(..., name="layer5"),
#   Concatenation("layer1", "layer5", 3, "layer6")
#   Conv2D(..., name="layer7"),
#   Conv2D(..., name="layer8"),
#   Dense(4, name="layer9")
# ], name="MyModel")
class SequentialWithConcatenations(Model):
    def __init__(layers, input=None, name=None):
        if input is None:
            input = layers.pop(0)
        output, lookup = input, dict()
        for index, layer in enumerate(layers):
            if isinstance(layer, Layer):
                output = layer(output)
                lookup[layer.name] = output
            else:
                assert isinstance(layer, Concatenation)
                layer_one, layer_two, axis, name = layer
                output = Concatenate(axis=axis, name=name)(
                    [lookup[layer_one], lookup[layer_two]]
                )
        super().__init__(input=input, output=output, name=name)


def get_specs_from_mosaic(mosaic):
    def get_spec_from_mosaic(mosaic, row, col):
        rows, cols = [], []
        for r in range(len(mosaic)):
            for c in range(len(mosaic[r])):
                if mosaic[r][c] == mosaic[row][col]:
                    rows.append(r)
                    cols.append(c)
        min_of_rows, max_of_rows = min(rows), max(rows)
        min_of_cols, max_of_cols = min(cols), max(cols)
        #
        if row == min_of_rows and col == min_of_cols:
            spec = {}
            if min_of_rows < max_of_rows:
                spec['rowspan'] = max_of_rows - min_of_rows + 1
            if min_of_cols < max_of_cols:
                spec['colspan'] = max_of_cols - min_of_cols + 1
            return spec
        return None

    if isinstance(mosaic, str):
        mosaic = [line.strip() for line in mosaic.splitlines()]
    mosaic = [[c for c in line] for line in mosaic if line]
    return [
        [
            get_spec_from_mosaic(mosaic, r, c)
            for c in range(len(mosaic[r]))
        ] for r in range(len(mosaic))
    ]



from PDSUtilities.tensorflow.view_model import view_model
from PDSUtilities.tensorflow.SequentialWithConcatenations import Concatenation
from PDSUtilities.tensorflow.SequentialWithConcatenations import SequentialWithConcatenations